#include "Preconditioner.h"

namespace AMDiS
{

  template<typename VectorType>
  BiCGStab<VectorType>::BiCGStab(::std::string name)
    : OEMSolver<VectorType>(name),
      r(NULL), rstar(NULL), p(NULL), v(NULL), t(NULL), xmin(NULL)
  {}

  template<typename VectorType>
  BiCGStab<VectorType>::~BiCGStab()
  {}

  template<typename VectorType>
  void BiCGStab<VectorType>::init()
  {
    r     = this->vectorCreator->create();
    rstar = this->vectorCreator->create();
    p     = this->vectorCreator->create();
    v     = this->vectorCreator->create();
    t   = this->vectorCreator->create();
    xmin  = this->vectorCreator->create();
  }

  template<typename VectorType>
  void BiCGStab<VectorType>::exit()
  {
    this->vectorCreator->free(r);
    this->vectorCreator->free(rstar);
    this->vectorCreator->free(p);
    this->vectorCreator->free(v);
    this->vectorCreator->free(t);
    this->vectorCreator->free(xmin);
  }

  template<typename VectorType>
  int BiCGStab<VectorType>::solveSystem(MatVecMultiplier<VectorType> *mv,
					VectorType *x, VectorType *b)
  {
    FUNCNAME("BiCGStab::solveSystem");

    double res, old_res = -1.0;
    double rh1, rh2, dad, alpha, ast, att, omega, beta;
    int    iter;

    const double TOL = 1e-30;

    /*------------------------------------------------------------------------*/
    /*---  Initalization  ----------------------------------------------------*/
    /*------------------------------------------------------------------------*/

    *p = *b;
    if (this->leftPrecon)
      this->leftPrecon->precon(p);
    double normB = norm(p);

    if (normB < TOL) {
      INFO(this->info,2)("b == 0; x = 0 is the solution of the linear system\n");
      setValue(*x, 0.0);
      this->residual = 0.0;
      return(0);
    }

    double save_tolerance = this->tolerance;
    if (this->relative)
      this->tolerance *= normB;

    *xmin = *x;
    int imin = 0;

    // h = Ax
    mv->matVec(NoTranspose, *x, *r);

    // h -= b
    *r *= -1.0;
    *r += *b;

    // preconditioning
    if (this->leftPrecon) this->leftPrecon->precon(r);

    // rstar = h;
    *rstar = *r;

    /*---  check initial residual  -------------------------------------------*/

    res = norm(r);

    START_INFO();
    if (SOLVE_INFO(0, res, &old_res)) {
      if (this->relative)
	this->tolerance = save_tolerance;
      return(0);
    }

    double normrmin = res;

    // d = h
    *p = *r;

    // rh1 = h*rstar
    rh1 = *r * *rstar;

    /*------------------------------------------------------------------------*/
    /*---  Iteration  --------------------------------------------------------*/
    /*------------------------------------------------------------------------*/

    for (iter = 1; iter <= this->max_iter; iter++) {
      /*---  calculate CA.d and CA.d*rstar  ----------------------------------*/
      
      // CAd = Ad
      mv->matVec(NoTranspose, *p, *v);
      
      if (this->leftPrecon) this->leftPrecon->precon(v);
      
      // dad = CAd * rstar;
      dad = *v * *rstar;
      
      alpha = rh1/dad;
      
      if (abs(alpha) > 1e+6 * (res/norm(v))) {
	BREAK_INFO("(Ad,d^*)_2 = 0", iter, res, &old_res);
	break;
      }

      /*---  update s and t  ------------------------------------------------*/
      
      // s = h
      
      // s = -alpha CAd + s
      axpy(-alpha, *v, *r);
      
      // x = -alpha d + x
      axpy(alpha, *p, *x);
      
      res = norm(r);
      if (res <= this->tolerance) {
	double red = res/old_res;
	INFO(this->info,2)("%5d | %12.5e | %8.2e\n", iter, res, red);
	INFO(this->info,6)("finished successfully with %d iterations\n", iter);
	if (this->relative)
	  this->tolerance = save_tolerance;
	return(iter);
      }

      if (res < normrmin) {    // update minimal norm quantities
	normrmin = res;
	*xmin    = *x;
	imin     = iter;
      }	else if (res > normrmin * 1e+6) {
	INFO(this->info,2)("Linear solver diverges.\n");
	INFO(this->info,2)("Current iteration: %d.\n", iter);
	INFO(this->info,2)("Current residual: %e.\n", res);
	break;
      }

      // t = As
      mv->matVec(NoTranspose, *r, *t);
      
      if (this->leftPrecon) 
	this->leftPrecon->precon(t);

      /*---  calculate omega  ------------------------------------------------*/
      
      // ast = s*t
      ast = *r * *t;
      
      // att = t*t
      att = *t * *t;
      
      omega = ast/att;
	
      if (abs(omega) <= 1e-6 * (res/norm(t))) {
	BREAK_INFO("omega = 0", iter, res, &old_res);
	break;
      }

      /*---   update x and calculate new h  ----------------------------------*/

      // x = -omega s + x
      axpy(omega, *r, *x);
      
      // h = s - omega * h
      axpy(-omega, *t, *r);
      
      res = norm(r);
      if (SOLVE_INFO(iter, res, &old_res) == 1) {
	if (this->relative)
	  this->tolerance = save_tolerance;
	return(iter);
      }

      if (res < normrmin) {    // update minimal norm quantities
	normrmin = res;
	*xmin    = *x;
	imin     = iter;
      } else if (res > normrmin * 1e+6) {
	INFO(this->info,2)("Linear solver diverges.\n");
	INFO(this->info,2)("Current iteration: %d.\n", iter);
	INFO(this->info,2)("Current residual: %e.\n", res);
	break;
      }

      // rh2 = h * rstar
      rh2 = *r * *rstar;
      
      beta = (rh2/rh1) * (alpha/omega);
      
      // d = -omega CAd + d
      axpy(-omega, *v, *p);
      
      // d = h + beta d
      *p *= beta;
      *p += *r;
      
      rh1 = rh2;
    }

    // returned solution is first with minimal residual
    *x       = *xmin;
    iter     = imin;
    this->residual = normrmin;

    if (this->relative) 
      this->tolerance = save_tolerance;

    INFO(this->info,2)("The minimal norm was %e; it was achieved in iteration %d.\n",
		       this->residual, iter);

    return(iter);
  }

}
